# Tai Sakuma <tai.sakuma@gmail.com>
import multiprocessing
import logging
# import logging.handlers

try:
    from logging.handlers import QueueHandler
except ImportError:
    from .queuehandler import QueueHandler

from alphatwirl import progressbar

##__________________________________________________________________||
class Worker(multiprocessing.Process):
    def __init__(self, task_queue, result_queue, logging_queue,
                 lock, progressReporter):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.logging_queue = logging_queue
        self.lock = lock
        self.progressReporter = progressReporter

    def run(self):
        self._configure_logger()
        self._configure_progressbar()
        try:
            self._run_tasks()
        except KeyboardInterrupt:
            pass

    def _configure_logger(self):
        handler = QueueHandler(self.logging_queue)
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def _configure_progressbar(self):
        progressbar._progress_reporter = self.progressReporter

    def _run_tasks(self):
        while True:
            message = self.task_queue.get()
            if message is None:
                self.task_queue.task_done()
                break
            task_idx, package = message
            result = package.task(*package.args, **package.kwargs)
            self.task_queue.task_done()
            self.result_queue.put((task_idx, result))


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

# The tests in this file are similar to those in test_selection.py.
# While __call__() is used in test_selection.py, the method event() is
# used in this file.
# The method event() is to be deleted, and so is this test file.
# The __call__() should be used. The selections are not specific to
# events. The method event() is kept for the comparability for now.

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules import All, Any, Not
from alphatwirl.selection.modules import AllwCount, AnywCount, NotwCount

##__________________________________________________________________||
all_classes = [All, AllwCount]
all_classe_ids = [c.__name__ for c in all_classes]

any_classes = [Any, AnywCount]
any_classe_ids = [c.__name__ for c in any_classes]

not_classes = [Not, NotwCount]
not_classe_ids = [c.__name__ for c in not_classes]

##__________________________________________________________________||
@pytest.mark.parametrize('Class', all_classes, ids=all_classe_ids)
def test_all(Class):
    sel1 = mock.Mock()
    sel2 = mock.Mock()

    obj = Class()
    obj.add(sel1)
    obj.add(sel2)

    event = mock.Mock()

    obj.begin(event)

    # need to set return_value each time because a selection is not
    # guaranteed to be called

    sel1.return_value = True
    sel2.return_value = True
    assert obj.event(event)

    sel1.return_value = True
    sel2.return_value = False
    assert not obj.event(event)

    sel1.return_value = False
    sel2.return_value = True
    assert not obj.event(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj.event(event)

    sel1.assert_called_with(event)
    sel2.assert_called_with(event)

    obj.end()

@pytest.mark.parametrize('Class', any_classes, ids=any_classe_ids)
def test_any(Class):
    sel1 = mock.Mock()
    sel2 = mock.Mock()

    obj = Class()
    obj.add(sel1)
    obj.add(sel2)

    event = mock.Mock()

    obj.begin(event)

    # need to set return_value each time because a selection is not
    # guaranteed to be called

    sel1.return_value = True
    sel2.return_value = True
    assert obj.event(event)

    sel1.return_value = True
    sel2.return_value = False
    assert obj.event(event)

    sel1.return_value = False
    sel2.return_value = True
    assert obj.event(event)

    sel1.return_value = False
    sel2.return_value = False
    assert not obj.event(event)

    sel1.assert_called_with(event)
    sel2.assert_called_with(event)

    obj.end()

##__________________________________________________________________||
@pytest.mark.parametrize('Class', all_classes, ids=all_classe_ids)
def test_all_empty(Class):
    obj = Class()
    event = mock.Mock()
    obj.begin(event)
    assert obj.event(event)
    obj.end()

@pytest.mark.parametrize('Class', any_classes, ids=any_classe_ids)
def test_any_empty(Class):
    obj = Class()
    event = mock.Mock()
    obj.begin(event)
    assert not obj.event(event)
    obj.end()

##__________________________________________________________________||
@pytest.mark.parametrize('Class', not_classes, ids=not_classe_ids)
def test_not(Class):
    sel1 = mock.Mock()
    sel1.side_effect = [True, False]

    obj = Class(sel1)

    event = mock.Mock()

    obj.begin(event)

    assert not obj.event(event)
    assert obj.event(event)

    obj.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from .list_to_aligned_text import list_to_aligned_text
from alphatwirl.misc.deprecation import _deprecated

##__________________________________________________________________||
@_deprecated(msg='use list_to_aligned_text() instead.')
def listToAlignedText(src, formatDict=None, leftAlignLastColumn=False):
    return list_to_aligned_text(
        src=src,
        format_dict=formatDict,
        left_align_last_column=leftAlignLastColumn
    )
##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

##__________________________________________________________________||
# copied from http://plumberjack.blogspot.co.uk/2010/09/using-logging-with-multiprocessing.html

class QueueHandler(logging.Handler):
    """
    This is a logging handler which sends events to a multiprocessing queue.

    The plan is to add it to Python 3.2, but this can be copy pasted into
    user code for use with earlier Python versions.
    """

    def __init__(self, queue):
        """
        Initialise an instance, using the passed queue.
        """
        logging.Handler.__init__(self)
        self.queue = queue

    def emit(self, record):
        """
        Emit a record.

        Writes the LogRecord to the queue.
        """
        try:
            ei = record.exc_info
            if ei:
                dummy = self.format(record) # just to get traceback text into record.exc_text
                record.exc_info = None  # not needed any more
            self.queue.put_nowait(record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

##__________________________________________________________________||




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2019-01-25T11:47:16+0000",
 "dirty": false,
 "error": null,
 "full-revisionid": "da6a460bd0c6ec38844e7471b0fa23792df0965e",
 "version": "0.20.3"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules.Count import Count

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return Count()

@pytest.fixture()
def sel1():
    ret= mock.Mock()
    ret.name ='sel1'
    ret.__class__.__name__ = 'MockEventSelection'
    return ret

@pytest.fixture()
def sel2():
    ret= mock.Mock()
    ret.name ='sel2'
    ret.__class__.__name__ = 'MockEventSelection'
    return ret

@pytest.fixture()
def sel3():
    ret= mock.Mock()
    ret.name = '' # empty string
    ret.__class__.__name__ = 'MockEventSelection'
    return ret

##__________________________________________________________________||
def test_add(sel1, sel2, sel3):
    obj = Count()
    assert 0 == len(obj._results)

    obj = Count(selections=[sel1, sel2])
    assert 2 == len(obj._results)
    obj.add(sel3)
    assert 3 == len(obj._results)

def test_repr(obj):
    repr(obj)

##__________________________________________________________________||
def test_copy(obj):
    obj._results[:] = [[1, 'class', 'name', 2, 3]]
    obj1 = obj.copy()
    assert obj is not obj1
    assert obj._results is not obj1._results
    assert [[1, 'class', 'name', 2, 3]] == obj1._results

def test_increment_depth(obj):
    obj._results[:] = [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
    ]

    obj.increment_depth(by=1)

    assert [
        [2, 'class1', 'name1', 6, 8],
        [2, 'class1', 'name2', 3, 6],
        [2, 'class2', 'name3', 1, 3],
    ] == obj._results

def test_insert(obj):
    obj._results[:] = [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
    ]

    obj1 = Count()
    obj1._results[:] = [
        [2, 'class2', 'name4', 4, 6],
        [2, 'class3', 'name5', 3, 4],
    ]

    obj.insert(1, obj1)

    assert [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [2, 'class2', 'name4', 4, 6],
        [2, 'class3', 'name5', 3, 4],
        [1, 'class2', 'name3', 1, 3],
    ] == obj._results

def test_insert_at_end(obj):

    obj._results[:] = [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
    ]

    obj1 = Count()
    obj1._results[:] = [
        [2, 'class2', 'name4', 2, 3],
        [2, 'class3', 'name5', 1, 2],
    ]

    obj.insert(2, obj1)

    assert [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
        [2, 'class2', 'name4', 2, 3],
        [2, 'class3', 'name5', 1, 2],
    ] == obj._results

##__________________________________________________________________||
def test_empty(obj):
    obj.count(pass_=[ ])

def test_one(obj, sel1):
    obj.add(sel1)

    assert [
        [1, 'MockEventSelection', 'sel1', 0, 0],
    ] == obj._results

    obj.count(pass_=[True])
    assert [
        [1, 'MockEventSelection', 'sel1', 1, 1],
    ] == obj._results

    obj.count(pass_=[False])
    assert [
        [1, 'MockEventSelection', 'sel1', 1, 2],
    ] == obj._results

def test_three(obj, sel1, sel2, sel3):
    obj.add(sel1)
    obj.add(sel2)
    obj.add(sel3)

    assert [
        [1, 'MockEventSelection', 'sel1', 0, 0],
        [1, 'MockEventSelection', 'sel2', 0, 0],
        [1, 'MockEventSelection',     '', 0, 0],
    ] == obj._results

    obj.count(pass_= [True, False])
    assert [
        [1, 'MockEventSelection', 'sel1', 1, 1],
        [1, 'MockEventSelection', 'sel2', 0, 1],
        [1, 'MockEventSelection',     '', 0, 0],
    ] == obj._results

    obj.count(pass_=[True, True, False])
    assert [
        [1, 'MockEventSelection', 'sel1', 2, 2],
        [1, 'MockEventSelection', 'sel2', 1, 2],
        [1, 'MockEventSelection',     '', 0, 1],
    ] == obj._results

def test_to_tuple_list(obj):
    obj._results[:] = [
        [1, 'class1', 'name1', 6, 8],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
    ]
    assert [
        (1, 'class1', 'name1', 6, 8),
        (1, 'class1', 'name2', 3, 6),
        (1, 'class2', 'name3', 1, 3)
    ] == obj.to_tuple_list()

##__________________________________________________________________||



from setuptools import setup, find_packages
import os
import versioneer

long_description = """
*AlphaTwirl* is a python library that summarizes event data into
multivariate categorical data as data frames. Event data, input to
AlphaTwirl, are data with one entry (or row) for one event: for
example, data in `ROOT <https://root.cern.ch/>`__ `TTrees
<https://root.cern.ch/doc/master/classTTree.html>`__ with one entry
per collision event of an `LHC
<https://home.cern/topics/large-hadron-collider>`__ experiment at
`CERN <http://home.cern/>`__. Event data are often large---too large
to be loaded in memory---because they have as many entries as events.
Multivariate categorical data, the output of AlphaTwirl, have one row
for one category. They are usually small---small enough to be loaded
in memory---because they only have as many rows as categories. Users
can, for example, import them as data frames into `R
<https://www.r-project.org/>`__ and `pandas
<http://pandas.pydata.org/>`__, which usually load all data in memory,
and can perform categorical data analyses with a rich set of data
operations available in R and pandas. """

setup(
    name='alphatwirl',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='A Python library for summarizing event data',
    long_description=long_description,
    author='Tai Sakuma',
    author_email='tai.sakuma@gmail.com',
    url='https://github.com/alphatwirl/alphatwirl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['docs', 'images', 'tests']),
)



# Tai Sakuma <tai.sakuma@gmail.com>

from .ReturnTrue import ReturnTrue

##__________________________________________________________________||
class PlusOne(object):
    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__,
        )

    def __call__(self, x):
        return x + 1

##__________________________________________________________________||
class Echo(object):
    def __init__(self, nextFunc=PlusOne(), valid=ReturnTrue()):
        self._nextFunc = nextFunc
        self._valid = valid

    def __repr__(self):
        return '{}(nextFunc={!r}, valid={!r})'.format(
            self.__class__.__name__,
            self._nextFunc,
            self._valid
        )

    def __call__(self, val):
        if not self._valid(val): return None
        return val

    def next(self, bin):
        if self._nextFunc is None: return None
        return self._nextFunc(bin)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class ProgressReport(object):
    """A progress report

    Args:
        name (str): the name of the task. if ``taskid`` is ``None``, used to identify the task
        done (int): the number of the iterations done so far
        total (int): the total iterations to be done
        taskid (immutable, optional): if given, used to identify the task. useful if multiple tasks have the same name
    """

    def __init__(self, name, done, total, taskid=None):
        self.taskid = taskid if taskid is not None else name
        self.name = name
        self.done = done
        self.total = total

    def __repr__(self):
        name_value_pairs = (
            ('taskid', self.taskid),
            ('name', self.name),
            ('done', self.done),
            ('total', self.total),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def last(self):
         return self.done == self.total

    def first(self):
         return self.done == 0

##__________________________________________________________________||



from alphatwirl.binning import Echo
import unittest

##__________________________________________________________________||
def plus2(val): return val + 2

##__________________________________________________________________||
class TestEcho(unittest.TestCase):
    def test_call(self):
        binning = Echo()
        self.assertEqual(1, binning(1))
        self.assertEqual(2, binning(2))
        self.assertEqual(0, binning(0))
        self.assertEqual(5, binning(5))

    def test_next_default(self):
        binning = Echo()
        self.assertEqual( -4, binning.next( -5))
        self.assertEqual(  1, binning.next(  0))
        self.assertEqual(  2, binning.next(  1))
        self.assertEqual(  4, binning.next(  3))
        self.assertEqual( 11, binning.next( 10))

    def test_next_plus2(self):
        binning = Echo(nextFunc = plus2)
        self.assertEqual( -3, binning.next( -5))
        self.assertEqual(  2, binning.next(  0))
        self.assertEqual(  3, binning.next(  1))
        self.assertEqual(  5, binning.next(  3))
        self.assertEqual( 12, binning.next( 10))

    def test_next_lambda(self):
        binning = Echo(nextFunc = lambda x: x + 0.1)
        self.assertEqual( -4.9, binning.next( -5))
        self.assertEqual(  0.1, binning.next(  0))
        self.assertEqual(  1.1, binning.next(  1))
        self.assertEqual(  3.1, binning.next(  3))
        self.assertEqual( 10.1, binning.next( 10))

    def test_next_None(self):
        binning = Echo(nextFunc = None)
        self.assertIsNone(binning.next( -5))
        self.assertIsNone(binning.next(  0))
        self.assertIsNone(binning.next(  1))
        self.assertIsNone(binning.next(  3))
        self.assertIsNone(binning.next( 10))

    def test_valid(self):
        binning = Echo(valid = lambda x: x >= 10)
        self.assertEqual( 13, binning(13))
        self.assertEqual( 10, binning(10))
        self.assertIsNone(binning(7))

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import inspect
import functools
import logging

##__________________________________________________________________||
def _removed(msg=''):
    def _decorate_class(c, msg):
        module_name = c.__module__
        logger = logging.getLogger(module_name)
        class_name = c.__name__
        name = '{}.{}'.format(module_name, class_name)
        text = '{} is removed.'.format(name)
        if msg:
            text += ' ' + msg

        def init(*args, **kwargs):
            logger.error(text)
            raise RuntimeError(msg)

        c.__init__ = init
        return c

    def _decorate_func(f, msg):
        module_name = f.__module__
        logger = logging.getLogger(module_name)
        name = f.__name__
        name = '{}.{}'.format(module_name, name)
        text = '{}() is removed.'.format(name)
        if msg:
            text += ' ' + msg

        @functools.wraps(f)
        def g(*args, **kwargs):
            logger.error(text)
            raise RuntimeError(msg)
        return g

    def _imp(f):
        if inspect.isclass(f):
            return _decorate_class(f, msg)
        return _decorate_func(f, msg)
    return _imp

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
class MockFile(object):
    pass

##__________________________________________________________________||
class MockTree(object):
    def __init__(self, entries=100):
        self.entries = entries
        self.ievent = -1
        self.branchstatus = [ ]
        self.getEntry_called = False
        self.directory = MockFile()
    def GetDirectory(self):
        return self.directory
    def GetEntries(self):
        return self.entries
    def GetEntry(self, ientry):
        self.getEntry_called = True
        if ientry < self.entries:
            nbytes = 10
            self.ievent = ientry
        else:
            nbytes = 0
            self.ievent = -1
        return nbytes

    def SetBranchStatus(self, bname, status):
        self.branchstatus.append((bname, status))

##__________________________________________________________________||
class MockBranch(object):
    pass

##__________________________________________________________________||
class MockBranchBuilder(object):
    def __init__(self):
        self.next = None
    def __call__(self, tree, name):
        return self.next

##__________________________________________________________________||
def test_mocktree():

    tree = MockTree(entries=3)
    assert isinstance(tree.GetDirectory(), MockFile)
    assert 3 == tree.GetEntries()

    assert -1 == tree.ievent

    nbytes = 10
    assert nbytes == tree.GetEntry(0)
    assert 0 == tree.ievent
    assert nbytes == tree.GetEntry(1)
    assert 1 == tree.ievent
    assert nbytes == tree.GetEntry(2)
    assert 2 == tree.ievent
    assert 0 == tree.GetEntry(3)
    assert -1 == tree.ievent

##__________________________________________________________________||
def test_repr():
    tree = MockTree()
    events = Events(tree)
    repr(events)

def test_init_branch_status():
    tree = MockTree()
    assert [ ] == tree.branchstatus
    events = Events(tree)
    events.buildBranch = MockBranchBuilder()
    assert [('*', 0)] == tree.branchstatus

def test_getattr():
    tree = MockTree()
    events = Events(tree)
    branchBuilder = MockBranchBuilder()
    events.buildBranch = branchBuilder

    branchBuilder.next = MockBranch()
    jet_pt = events.jet_pt
    assert isinstance(jet_pt, MockBranch)

def test_getattr_same_objects():
    tree = MockTree()
    events = Events(tree)
    branchBuilder = MockBranchBuilder()
    events.buildBranch = branchBuilder

    branch1 = MockBranch()
    branchBuilder.next = branch1
    jet_pt1 = events.jet_pt
    assert branch1 is jet_pt1

    branch2 = MockBranch()
    branchBuilder.next = branch2
    jet_pt2 = events.jet_pt

    assert branch1 is jet_pt2
    assert branch2 is not jet_pt2

    it = iter(events)
    event = next(it)

    branch3 = MockBranch()
    branchBuilder.next = branch3
    jet_pt3 = event.jet_pt

    assert branch1 is jet_pt3
    assert branch3 is not jet_pt3

def test_getattr_exception():
    tree = MockTree()
    events = Events(tree)
    events.buildBranch = MockBranchBuilder()

    with pytest.raises(AttributeError):
        events.no_such_branch

def test_getattr_getentry():
    tree = MockTree()
    events = Events(tree, start=10)
    branchBuilder = MockBranchBuilder()
    events.buildBranch = branchBuilder

    assert -1 == events.iEvent
    assert -1 == tree.ievent
    assert not tree.getEntry_called
    branchBuilder.next = MockBranch()
    jet_pt = events.jet_pt
    assert not tree.getEntry_called

    it = iter(events)
    event = next(it)
    tree.getEntry_called = False
    assert 0 == events.iEvent
    assert 10 == tree.ievent
    assert not tree.getEntry_called
    jet_pt = event.jet_pt
    assert not tree.getEntry_called

    tree.getEntry_called = False
    met_pt = event.met_pt
    assert tree.getEntry_called
    assert 0 == events.iEvent
    assert 10 == tree.ievent

    event = next(it)
    tree.getEntry_called = False
    assert 1 == events.iEvent
    assert 11 == tree.ievent
    assert not tree.getEntry_called
    jet_pt = event.jet_pt
    assert not tree.getEntry_called

    tree.getEntry_called = False
    met_pt = event.met_pt
    assert not tree.getEntry_called
    assert 1 == events.iEvent
    assert 11 == tree.ievent

    tree.getEntry_called = False
    njets = event.njets
    assert tree.getEntry_called
    assert 1 == events.iEvent
    assert 11 == tree.ievent

def test_pass_arguments_to_base_class():
    tree = MockTree(entries=1000)
    events = Events(tree, 100)
    assert 100 == events.nEvents
    assert 0 == events.start

    events = Events(tree, 100, 20)
    assert 100 == events.nEvents
    assert 20 == events.start

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import EventBuilderConfig
    from alphatwirl.roottree import BEventBuilder

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def obj(request):
    config = EventBuilderConfig(
        inputPaths=['/path/to/input1/tree.root', '/path/to/input2/tree.root'],
        treeName='tree',
        maxEvents=123,
        start=11,
        name='TTJets'
    )
    return BEventBuilder(config)

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

##__________________________________________________________________||



import unittest

import copy
import numpy as np

from alphatwirl.summary import Scan

##__________________________________________________________________||
class TestScan(unittest.TestCase):

    def test_init(self):
        obj = Scan()
        np.testing.assert_equal([ ], obj.contents)

    def test_init_val(self):
        obj = Scan(val = (10, 20))
        self.assertEqual([(10, 20)], obj.contents)

    def test_init_weight(self): # no effect
        obj = Scan(val = (10, 20), weight = 2)
        self.assertEqual([(10, 20)], obj.contents)

    def test_init_contents(self):
        obj = Scan(contents = [(10, 20), (30, 40)])
        self.assertEqual([(10, 20), (30, 40)], obj.contents)

    @unittest.skip("they can be the same object now")
    def test_init_contents_not_same_object(self):
        contents = [[10, 20], [30, 40]]
        obj = Scan(contents = contents)
        self.assertIsNot(contents, obj.contents)
        self.assertIsNot(contents[0], obj.contents[0])
        self.assertIsNot(contents[1], obj.contents[1])

    def test_repr(self):
        obj = Scan()
        repr(obj)

    def test_add(self):
        obj1 = Scan(contents = [(10, 20), (30, 40)])
        obj2 = Scan(contents = [(50, 60), (70, 80)])
        obj3 =  obj1 + obj2
        self.assertEqual([(10, 20), (30, 40), (50, 60), (70, 80)], obj3.contents)
        self.assertIsNot(obj1, obj3)
        self.assertIsNot(obj1.contents, obj3.contents)
        self.assertIsNot(obj2, obj3)
        self.assertIsNot(obj2.contents, obj3.contents)

    def test_radd(self):
        obj1 = Scan(contents = [(10, 20), (30, 40)])
        self.assertIsNot(obj1, sum([obj1])) # will call 0 + obj1
        self.assertEqual(obj1, sum([obj1]))

    def test_radd_raise(self):
        obj1 = Scan(contents = [(10, 20), (30, 40)])
        self.assertRaises(TypeError, obj1.__radd__, 1)

    def test_copy(self):
        obj1 = Scan(contents = [(10, 20), (30, 40)])
        copy1 = copy.copy(obj1)
        self.assertEqual(obj1, copy1)
        self.assertIsNot(obj1, copy1)
        self.assertIsNot(obj1.contents, copy1.contents)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from .Events import Events
from .BranchBuilder import BranchBuilder

##__________________________________________________________________||
class BEvents(Events):
    def __init__(self, tree, maxEvents=-1, start=0):
        super(BEvents, self).__init__(tree, maxEvents, start)
        self.branches = { }
        self.buildBranch = BranchBuilder()
        self.buildBranch.register_tree(tree)

    def __repr__(self):
        return '{}({}, branches={!r})'.format(
            self.__class__.__name__,
            super(BEvents, self)._repr_contents(),
            self.branches
        )

    def __getattr__(self, name):

        if name in self.branches:
            return self.branches[name]

        branch = self.buildBranch(self.tree, name)

        if branch is None:
            raise AttributeError('{!r} has no attribute "{}"'.format(self, name))

        self.branches[name] = branch

        if self.iEvent >= 0:
            self.tree.GetEntry(self.start + self.iEvent)

        return self.branches[name]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from .expand import expand_path_cfg

##__________________________________________________________________||
def AllFactory(components, name=None, **kargs):
    ret = kargs['AllClass'](name=name)
    for c in components:
        ret.add(c)
    return ret

def AnyFactory(components, name=None,  **kargs):
    ret = kargs['AnyClass'](name=name)
    for c in components:
        ret.add(c)
    return ret

def NotFactory(components, name=None, **kargs):
    return kargs['NotClass'](selection=components[0], name=name)

##__________________________________________________________________||
def LambdaStrFactory(lambda_str, LambdaStrClass, name=None, **kargs):
    return LambdaStrClass(lambda_str=lambda_str.format(**kargs), name=name)

##__________________________________________________________________||
def FactoryDispatcher(path_cfg, **kargs):

    alias_dict = kargs.get('aliasDict', { })
    path_cfg = expand_path_cfg(path_cfg, alias_dict=alias_dict)

    return call_factory(path_cfg, **kargs)

##__________________________________________________________________||
def call_factory(path_cfg, **kargs):
    components = [call_factory(c, **kargs) for c in path_cfg['components']]
    path_cfg_copy = path_cfg.copy()
    factoryName = path_cfg_copy.pop('factory')
    factory = find_factory(factoryName)
    kargs_copy = kargs.copy()
    kargs_copy.update(path_cfg_copy)
    kargs_copy['components'] = components
    return factory(**kargs_copy)

##__________________________________________________________________||
def find_factory(name):

    ret_dict = dict(
        AllFactory=AllFactory,
        AnyFactory=AnyFactory,
        NotFactory=NotFactory,
        LambdaStrFactory=LambdaStrFactory,
    )

    return ret_dict[name]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

from .Branch import Branch
from .BranchAddressManager import BranchAddressManager
from .BranchAddressManagerForVector import BranchAddressManagerForVector

from .inspect import is_ROOT_null_pointer

##__________________________________________________________________||
branchAddressManager = BranchAddressManager()
branchAddressManagerForVector = BranchAddressManagerForVector()

##__________________________________________________________________||
class BranchBuilder(object):
    """This class builds a branch.

    A branch is an object with data whose address is set in a tree.

    All instances of this class share the data.

    """

    itsdict = { }

    def __repr__(self):
        name_value_pairs = (
            ('itsdict', self.__class__.itsdict),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def register_tree(self, tree):
        if not tree in self.__class__.itsdict:
            tree.SetBranchStatus('*', 0)
            self.__class__.itsdict[tree] = { }

    def __call__(self, tree, name):

        if not tree in self.__class__.itsdict:
            logger = logging.getLogger(__name__)
            logger.warning('tree is not registered: {!r}'.format(tree))
            self.__class__.itsdict[tree] = { }

        itsdict_tree = self.__class__.itsdict[tree]
        if name in itsdict_tree:
            return itsdict_tree[name]

        branch = self.imp(tree, name)
        itsdict_tree[name] = branch
        return branch

    def imp(self, tree, name):
        if not self._branch_exist(tree, name): return None

        branch = self._try_ctypes_or_array_of_ctypes(tree, name)
        if branch is not None: return branch

        branch = self._try_std_vector(tree, name)
        if branch is not None: return branch

        self._unknown_type_warning(tree, name)

        return None

    def _branch_exist(self, tree, name):
        leaves = tree.GetListOfLeaves()

        try:
            leafNames = [l.GetName() for l in leaves]
        except TypeError:
            logger = logging.getLogger(__name__)
            logger.warning(
                'cannot get leaf names of the tree: '
                '{!r}'.format(tree))
            if is_ROOT_null_pointer(leaves):
                logger.warning(
                    'tree.GetListOfLeaves() returns null pointer. '
                    'the tree might be a TChain with no files.')
            return False

        if name in leafNames:
            return True
        return False

    def _try_ctypes_or_array_of_ctypes(self, tree, name):
        itsArray, itsCountArray = branchAddressManager.getArrays(tree, name)
        if itsArray is None: return None
        branch = Branch(name, itsArray, itsCountArray)
        return branch

    def _try_std_vector(self, tree, name):
        itsVector = branchAddressManagerForVector.getVector(tree, name)
        return itsVector # this can be used as a branch

    def _unknown_type_warning(self, tree, name):
        leaf = tree.GetLeaf(name)
        typename = leaf.GetTypeName()
        logger = logging.getLogger(__name__)
        logger.warning('unknown leaf type : {}'.format(typename))

##__________________________________________________________________||
